"""
Convenience layer module for integration using multiple level sets.
"""
from ngsolve import Norm, Grad, GridFunction, CoefficientFunction, IfPos
from xfem import IF, POS, NEG, ANY, Integrate
from itertools import chain, product, repeat
from collections import Counter


class DomainTypeArray():
    """
    Container class for multiple level sets.

    Parameters
    ----------
    dtlist : {list(tuples({ENUM,ANY})), tuple({ENUM,ANY})}
        The tuple(s) describing the region of interest. The regions
        described by the individual tuples must be of the same
        co-dimension and the same length. Each ANY is expanded
        internally into a POS and a NEG region.
    lsets : tuple(ngsolve.GridFunction)
        Optional argument (default=None): A tuple of discrete level set
        functions, with respect to which the domain in defined. If this
        given, then the DomainTypeArray is compressed into regions with
        positive measure on initialisation.
    persistent_compress : bool
        Optional argument (default=False): Sets whether the result from
        operations on this DomainTypeArray instance should automatically
        be compressed with respect to lsets.

    Attributes
    ----------
    as_list : list(tuples)
        Expanded list containing tuples describing the region of
        interest.
    codim : int
        Co-dimension of the contained region of interest.
    lsets : {None, tuple(ngsolve.GridFunction)}
        The set of level set functions, with respect to which the
        instance is compressed if persistent_compress=True.
    persistent_compress : bool
        Indicates, whether the instance to be compressed persistently
        after operations. the default value is False.
    do_safety_checks : bool
        Option to turn of safety checks on the input list. This is
        useful if a copy of an existing DomainTypeArray is made and it
        is known, that the input is permissible. Default is value is
        True.

    Methods
    -------
    Boundary(element_marking=False):
        A DomainTypeArray describing the boundary of the current
        DomainTypeArray region.
    Compress(lsets, persistent=False):
        Remove domain regions which have have zero measure with respect
        to the given level set functions.
    Indicator(lsets):
        Returns an indicator CoefficientFunction of the instances'
        region (codim=0 only).
    IndicatorSmoothed(lsets, eps):
        Returns an indicator CoefficientFunction of an eps-region 
        around the current instances' region (codim>0 only).
    GetOuterNormals(lsetsp1):
        Returns a dictionary, containing the outward pointing unit
        normal vectors on each section of self.Boundary(). The keys are
        the tuples of the boundary segments, each normal vector is
        defined on.
    """

    def __init__(self, dtlist, lsets=None, persistent_compress=False,
                 do_safety_checks=True):
        if type(dtlist) == tuple:
            dtlist = [dtlist]
        if type(dtlist) != list or type(dtlist[0]) != tuple:
            raise TypeError("Invalid input: dtlist must be a tuple or a list of tuples")

        dtt_len = len(dtlist[0])

        self.codim = sum([1 for dt in dtlist[0] if dt == IF])
        if self.codim > 3:
            print("Warning: Codim > 3 !!!")
        if do_safety_checks:
            for dtt in dtlist:
                for dt in dtt:
                    if dt not in [NEG, POS, IF, ANY]:
                        raise Exception("Initialised with something other than NEG"
                                        ", POS, IF, ANY")
                n = dtt.count(IF)
                if n != self.codim:
                    raise Exception("DomainTypeArray initialised with tuples of "
                                    "different co-dimension!")
                if dtt_len != len(dtt):
                    raise Exception("DomainTypeArray initialised with tuples of "
                                    "different length!")
            if lsets:
                if len(lsets) != dtt_len:
                    raise Exception("The number of level sets does not match the "
                                    " length of the arrays domain tuples!")
                if type(lsets[0]) != GridFunction:
                    raise Exception("The level set functions need to be "
                                    "ngsolve.GridFunctions!")

        # Expansion of ANY
        if (ANY in chain.from_iterable(dtlist)):
            dtlist_expand = []
            for dtt in dtlist:
                if ANY in dtt:
                    locations_expand = [i for i in range(len(dtt)) if dtt[i] == ANY]
                    n = len(locations_expand)
                    dt_lists = [list(dtt) for i in range(2**n)]
                    dt_per = list(product(*list(repeat((POS, NEG), n))))
                    for i, dtl in enumerate(dt_lists):
                        for j, k in enumerate(locations_expand):
                            dt_lists[i][k] = dt_per[i][j]
                        dtlist_expand.append(tuple(dt_lists[i]))
                else:
                    dtlist_expand.append(dtt)
            self.as_list = list(set(dtlist_expand))
        else:
            self.as_list = list(set(dtlist))

        self.lsets = lsets
        self.persistent_compress = persistent_compress

        if self.lsets:
            self.Compress(self.lsets, self.persistent_compress)
        
        if not self.persistent_compress:
            self.lsets = None

    def __len__(self):
        return self.as_list.__len__()

    def __iter__(self):
        return self.as_list.__iter__()

    def __contains__(self, dtt):
        return dtt in self.as_list

    def __eq__(self, dta_b):
        if self.codim != dta_b.codim or self.__len__() != len(dta_b):
            return False
        else:
            return Counter(self.as_list) == Counter(dta_b.as_list)

    def __ne__(self, dta_b):
        return not self.__eq__(dta_b)

    def __str__(self):
        str_out = "DomainTypeArray:\n"
        str_out += "  Co-dim : {}\n".format(self.codim)
        str_out += "  Domains: ["
        for dtt in self.as_list:
            str_out += dtt.__str__() + ", "
        return str_out[:-2] + "]"

    def __or__(self, dta_b):
        if self.codim != dta_b.codim:
            raise Exception("Co-dims don't match: Union not possible!")
        if not self.persistent_compress and not dta_b.persistent_compress:
            lsets_out = None
            pers_out = False
        elif ((self.persistent_compress and  not dta_b.persistent_compress) or 
              (not self.persistent_compress and dta_b.persistent_compress)):
            print("WARNING: Combined domains with non matching persistent "
                  "compression status: \n The returned domain will not be "
                  "compressed.")
            lsets_out = None
            pers_out = False
        else:
            for lset1, lset2 in zip(self.lsets, dta_b.lsets):
                if lset1 != lset2:
                    raise Exception("Operator only possible for domains with the same level sets")
            lsets_out = self.lsets
            pers_out = self.persistent_compress

        return DomainTypeArray(self.as_list + dta_b.as_list, lsets_out, pers_out)

    def __and__(self, dta_b):
        if not self.persistent_compress and not dta_b.persistent_compress:
            lsets_out = None
            pers_out = False
        elif ((self.persistent_compress and  not dta_b.persistent_compress) or 
              (not self.persistent_compress and dta_b.persistent_compress)):
            print("WARNING: Combined domains with non matching persistent "
                  "compression status: \n The returned domain will not be "
                  "compressed.")
            lsets_out = None
            pers_out = False
        else:
            for lset1, lset2 in zip(self.lsets, dta_b.lsets):
                if lset1 != lset2:
                    raise Exception("Operator only possible for domains with the same level sets")
            lsets_out = self.lsets
            pers_out = self.persistent_compress

        dtl_out = [dtt for dtt in self.as_list if dtt in dta_b.as_list]
        
        if len(dtl_out) == 0:
            for dtt1, dtt2 in product(self.as_list, dta_b.as_list):
                if dtt1 == dtt2:
                    dtl_out.append(dtt1)
                    continue
                dtt_out = []
                for i, dt in enumerate(dtt1):
                    if dt != dtt2[i]:
                        if dt in [POS, NEG] and dtt2[i] in [POS, NEG]:
                            break
                        else:
                            dtt_out.append(IF)
                    else:
                        dtt_out.append(dt)
                else:
                    dtl_out.append(tuple(dtt_out))
        
        return DomainTypeArray(dtl_out, lsets_out, pers_out)

    def __invert__(self):
        n = len(self.as_list[0])
        if self.codim == 0:
            dt_per = set(product(*list(repeat((POS, NEG), n))))
        else:
            dt_per = set(product(*list(repeat((POS, NEG, IF), n))))
            dt_per = set(filter(lambda dtt: dtt.count(IF) == self.codim, dt_per))
        dt_per = list(dt_per.difference(set(self.as_list)))
        return DomainTypeArray(dt_per, self.lsets, self.persistent_compress)

    def __ior__(self, dta_b):
        return self.__or__(dta_b)

    def __iand__(self, dta_b):
        return self.__and__(dta_b)

    def Compress(self, lsets, persistent=False):
        """
        For a given set of level sets, this function removes any regions
        contained in the instance, which do not have a positive measure.
        The level sets remain attached to the instance, such that 
        self.Boundary() called after self.Compress(lsets) automatically
        compresses the output. 
        
        Parameters
        ----------
        lsets : tuple(ngsolve.GridFunction)
            The set of level sets with respect to which it is computed,
            whether regions have positive measure.
        persistent : bool
            Sets compression to continue on operations such as 
            Boundary(). Default vale = False 

        Returns
        -------
        None
        """

        if len(lsets) != len(self.as_list[0]):
            raise Exception("The number of level sets does not match the "
                            " length of the arrays domain tuples!")
        if type(lsets[0]) != GridFunction:
            raise Exception("The level set functions need to be "
                            "ngsolve.GridFunctions!")

        mesh = lsets[0].space.mesh
        dtl_out = []
        for dtt in self.as_list:
            weight = Integrate({"levelset": lsets, "domain_type": dtt},
                               cf=1, mesh=mesh, order=0)
            if abs(weight) > 1e-12:
                dtl_out.append(dtt)
        self.as_list = dtl_out

        if persistent:
            self.persistent_compress = persistent
            self.lsets = lsets

        return None


    def Boundary(self):
        """
        Compute the boundary of the instances' contained region. If
        persistent_compress=True has been set for the instance, then
        the result will compressed into boundary regions with positive
        measure. 

        Returns
        -------
        DomainTypeArray
            Container of the resulting boundary region.
        """
        if self.codim >= 3:
            raise Exception("Boundary does not make sense for codim >=3")

        dtl_out = []
        for dtt in self.as_list:
            for i, dt in enumerate(dtt):
                if dt != IF:
                    if (dtt[:i] + (POS,) + dtt[i+1:] in self and
                        dtt[:i] + (NEG,) + dtt[i+1:] in self):
                        continue
                    else:
                        is_relevant = True
                        dtt_out = dtt[:i] + (IF,) + dtt[i+1:]
                        if self.lsets:
                            weight = Integrate({"levelset": self.lsets,
                                                "domain_type": dtt_out},
                                                cf=1, 
                                                mesh=self.lsets[0].space.mesh,
                                                order=0)
                            if abs(weight) < 1e-12:
                                is_relevant = False
                        if is_relevant:
                            dtl_out.append(dtt_out)

        if self.persistent_compress:
            dta_out = DomainTypeArray(dtl_out, self.lsets, 
                                      self.persistent_compress)
        else:
            dta_out = DomainTypeArray(dtl_out)
    
        return dta_out

    def Indicator(self, lsets):
        """
        Indicator CoefficientFunction for a DomainTypeArray of codim=0.

        Parameters
        ----------
        lsets : tuple(ngsolve.CoefficientFunctions)
            The set of level set functions with respect to which the
            instances' region is defined.

        Returns
        -------
        CoefficientFunction
            1 in the region on interest, 0 else.
        """

        if self.codim != 0:
            print("Warning: Indicator is not intended for codim > 0! "
                  "Please use the IndicatorSmoothed method")
            return self.IndicatorSmoothed(lsets)

        ind_combined = CoefficientFunction(0)
        for dtt in self.as_list:
            ind = CoefficientFunction(1)
            for i, dt in enumerate(dtt):
                if dt == POS:
                    ind *= IfPos(lsets[i], 1, 0)
                elif dt == NEG:
                    ind *= IfPos(-lsets[i], 1, 0)

            ind_combined += ind
            del ind

        ind_leveled = IfPos(ind_combined, 1, 0)
        del ind_combined

        return ind_leveled

    def IndicatorSmoothed(self, lsets, eps=0.03):
        """
        Indicator CoefficientFunction, indicating the eps-region around
        the instances' region (codim>0). We assume that the level sets
        are approximately signed distance functions in the eps region of
        the zero set. 

        Parameters
        ----------
        lsets : tuple(ngsolve.CoefficientFunctions)
            The set of level set functions defining the region.
        eps : float
            The distance around the sub-domain which is indicated
            (default eps=0.01).

        Returns
        -------
        CoefficientFunction
            1 in the region of distance eps around the sub-domain, 0 
            else.
        """

        if self.codim == 0:
            print("Warning: IndicatorSmothed is not intended for codim==0! "
                  "Please use the Indicator method")
            return self.Indicator(lsets)

        def gf_abs(gf):
            return IfPos(gf, gf, -gf)

        ind_combined = CoefficientFunction(0)
        for dtt in self.as_list:
            ind = CoefficientFunction(1)
            ind_cut = CoefficientFunction(1)

            for i, dt in enumerate(dtt):
                if dt == IF:
                    ind *= IfPos(gf_abs(lsets[i]) - eps, 0, 1)
                elif dt == POS:
                    ind_cut *= IfPos(lsets[i] - eps, 1, 0)
                elif dt == NEG:
                    ind_cut *= IfPos(-lsets[i] + eps, 1, 0)

            ind_combined += ind * ind_cut
            del ind, ind_cut

        ind_leveled = IfPos(ind_combined, 1, 0)

        del ind_combined
        return ind_leveled

    def GetOuterNormals(self, lsetsp1):
        """
        For each region in self.Boundary(), we compute the outward 
        pointing unit normal vector.

        Parameters
        ----------
        lesetsp1 : tuple(ngsolve.GridFunction)
            The set of P1 level set functions.

        Returns
        -------
        dict(ngsolve.GridFunction)
            Each normal is accessible with the tuple defining the
            boundary region it is defined on as the key.
        """

        if self.codim > 0:
            raise NotImplemented("GetOuterBoundary is olny available for "
                                 "codim = 0 !")

        n_dtt = {}
        bnd = self.Boundary()
        for dtt in self.as_list:
            for i, dt in enumerate(dtt):
                dtt_bnd = dtt[:i] + tuple([IF]) + dtt[i+1:]
                if dtt_bnd not in bnd:
                    continue
                if dt == POS:
                    n_dt = - 1.0 / Norm(Grad(lsetsp1[i])) * Grad(lsetsp1[i])
                else:
                    n_dt = 1.0 / Norm(Grad(lsetsp1[i])) * Grad(lsetsp1[i])
                n_dtt[dtt_bnd] = n_dt

        del bnd
        return n_dtt


def TensorUnion(*args):
    """
    Construct the union of an arbitrary number of regions, described by
    different DomainTypeArrays, each dependent on a different set of 
    level sets. This increases the tuple-dimension of the resulting 
    domain regions.

    Parameters
    ----------
    *args : DomainTypeArray
        Variable list of DomainTypeArray objects of the same
        co-dimension.

    Returns
    -------
    DomainTypeArray
        Container of the resulting region.
    """

    n_dtas = []
    codim = args[0].codim

    for dta in args:
        if type(dta) != DomainTypeArray:
            raise TypeError("TensorUnion only possible for DomainTypeArrays")
        if dta.codim != codim:
            raise Exception("Cannot form TensorUnion for arrays of different codimension")
        n_dtas.append(len(dta.as_list[0]))

    i = 0
    dtl_out = []
    dtt_candiate = [ANY for j in range(sum(n_dtas))]
    for j, dta in enumerate(args):
        for dtt in dta.as_list:
            dtt_out = dtt_candiate.copy()
            dtt_out[i:i+n_dtas[j]] = dtt   
            dtl_out.append(tuple(dtt_out))
        i += n_dtas[j]

    return DomainTypeArray(dtl_out)


def TensorIntersection(*args):
    """
    Construct the intersection of an arbitrary number of regions, 
    described by different DomainTypeArrays, each dependent on a
    different set of level sets. This increases the tuple-dimension of
    the resulting domain regions.

    Parameters
    ----------
    *args : DomainTypeArray
        Variable list of DomainTypeArray objects.

    Returns
    -------
    DomainTypeArray
        Container of the resulting region.
    """
    
    for dta in args:
        if type(dta) != DomainTypeArray:
            raise TypeError("TensorUnion only possible for DomainTypeArrays")

    dtl_out = []
    for dtt in product(*[dta.as_list for dta in args]):
        dtl_out.append(tuple(chain(*dtt)))
    
    return DomainTypeArray(dtl_out)



"""
(ngs)xfem
=========

A module for unfitted finite element discretizations in NGSolve

Submodules:
xfem.cutmg ... MultiGrid for CutFEM
xfem.lsetcurving ... isoparametric unfitted FEM
xfem.lset_spacetime ... isoparametric unfitted space-time FEM
xfem.mlset ... multiple level sets
xfem.utils ... some example level set geometries
"""



from ngsolve import (L2, VOL, BitArray, CoefficientFunction, FESpace,
                     GridFunction, H1, IfPos, LinearForm, Parameter, dx)
from ngsolve.comp import Integrate as ngsolve_Integrate
from ngsolve.comp import ProxyFunction
from ngsolve.comp import SymbolicBFI as ngsolve_SymbolicBFI
from ngsolve.comp import SymbolicLFI as ngsolve_SymbolicLFI
from xfem.ngsxfem_py import *
from xfem.ngs_check import check_if_ngsolve_newer_than, __ngsolve_required__

check_if_ngsolve_newer_than(__ngsolve_required__)

def HAS(domain_type):
    """
For a given domain_type return the combined domain type that 
includes all elements that have a part in the domain type.
    """
    if domain_type == NEG:
        return HASNEG
    elif domain_type == POS:
        return HASPOS
    elif domain_type == IF:
        return IF
    else:
        raise Exception("invalid domain type")

def extend(func):
    """
Evaluates the XFiniteElement-function independent of the level set domains.

Note:
This will lead to the same behavior as the function that the XFiniteElement-function is based
on.
    """
    if func.derivname == "extend":
        return func.Deriv()
    add = func.Operator("extend")
    if add:
        return add
    raise Exception("cannot form extend")

def pos(func):
    """
Evaluates an XFiniteElement-function assuming a positive level set domain.

Note:
This can lead to non-zero values also in domains where the level set function is non-positive.
    """
    if func.derivname == "pos":
        return func.Deriv()
    add = func.Operator("pos")
    if add:
        return add
    raise Exception("cannot form pos")

def neg(func):
    """
Evaluates an XFiniteElement-function assuming a negative level set domain.

Note:
This can lead to non-zero values also in domains where the level set function is non-negative.
    """
    if func.derivname == "neg":
        return func.Deriv()
    add = func.Operator("neg")
    if add:
        return add
    raise Exception("cannot form neg")

def extend_grad(func):
    """
Evaluates the gradient of an XFiniteElement-function independent of the level set domains.

Note:
This will lead to the same behavior as the function that the XFiniteElement-function is based on.
    """
    if func.derivname == "extendgrad":
        return func.Deriv()
    add = func.Operator("extendgrad")
    if add:
        return add
    raise Exception("cannot form extend_grad")

def pos_grad(func):
    """
Evaluates the gradient of an XFiniteElement-function assuming a positive level set domain.

Note:
This can lead to non-zero values also in domains where the level set function is non-positive.
    """
    if func.derivname == "posgrad":
        return func.Deriv()
    add = func.Operator("posgrad")
    if add:
        return add
    raise Exception("cannot form pos_grad")

def neg_grad(func):
    """
Evaluates the gradient of an XFiniteElement-function assuming a negative level set domain.

Note:
This can lead to non-zero values also in domains where the level set function is non-negative.
    """
    if func.derivname == "neggrad":
        return func.Deriv()
    add = func.Operator("neggrad")
    if add:
        return add
    raise Exception("cannot form neg_grad")

def dtref(func):
    """
Evaluates the time derivative (w.r.t. the reference time interval) of a Space-Time function.
    """
    add = func.Operator("dt")
    if add:
        return add
    raise Exception("cannot form dt")

def dt(func):
    """
Deprecated: use "dtref" instead
    """
    print("WARNING: dt is deprecated. Use \"dtref\" instead. \n         Note that the operator acts w.r.t. the reference time intervals.")
    return dtref(func)



def SymbolicBFIWrapper(levelset_domain=None, *args, **kwargs):
    """
Wrapper around SymbolicBFI to allow for integrators on level set domains (see also
SymbolicCutBFI). The dictionary contains the level set function (CoefficientFunciton or
GridFunction) and the domain-type (NEG/POS/IF). If the dictionary is not provided, the standard
SymbolicBFI function from NGSolve will be called.

Parameters

levelset_domain : dictionary
  entries:
  * "levelset": 
    singe level set : ngsolve.CoefficientFunction
      CoefficientFunction that describes the geometry. In the best case lset is a GridFunction of an
      FESpace with scalar continuous piecewise (multi-) linear basis functions.
    multiple level sets: tuple(ngsolve.GridFunction)
      Tuple of GridFunctions that describe the geometry.
  * "domain_type" :
    single level set: {NEG,POS,IF} (ENUM) 
      Integration on the domain where either:
      * the level set function is negative (NEG)
      * the level set function is positive (POS)
      * the level set function is zero     (IF )
    multiple level sets: {tuple({ENUM}), list(tuple(ENUM)), DomainTypeArray}
      Integration on the domains specified
  * "subdivlvl" : int
    On simplex meshes a subtriangulation is created on which the level set function lset is
    interpolated piecewise linearly. Based on this approximation, the integration rule is
    constructed. Note: this argument only works on simplices.
  * "order" : int
    (default: entry does not exist or value -1)
    overwrites "order"-arguments in the integration
  * "quad_dir_policy" : {FIRST, OPTIMAL, FALLBACK} (ENUM)
    Integration direction policy for iterated integrals approach
    * first direction is used unless not applicable (FIRST)
    * best direction (in terms of transformation constant) is used (OPTIMAL)
    * subdivision into simplices is always used (FALLBACK)

Other Parameters :

  form : ngsolve.CoefficientFunction
    form to integrate

  VOL_or_BND : {VOL,BND}
    integrator is defined in the volume or boundary

  element_boundary : boolean
    Integration of the boundary of an element
    (not active for level set domains)

  skeleton : boolean
    Integration over element-interface

  definedon : Region
    Domain description on where the integrator is defined

  definedonelements: BitArray
    BitArray that allows integration only on elements or facets (if skeleton=True) that are marked
    True.

  deformation : GridFunction
    Specify a specific mesh deformation for a bilinear form

  order : int
    Modifies the order of the quadrature rule used. This is overruled by "order"-entry of the 
    levelset_domain dictionary, if the dictionary entry exists.

  time_order : int
    order in time that is used in the space-time integration. time_order=-1 means that no space-time
    rule will be applied. This is only relevant for space-time discretizations.
"""
    if levelset_domain != None and type(levelset_domain)==dict:
        # shallow copy is sufficient to modify "order" and "time_order" locally
        levelset_domain_local = levelset_domain.copy()
        if "order" in kwargs:
            if not "order" in levelset_domain_local:
                levelset_domain_local["order"] = kwargs["order"]
            del kwargs["order"]
        if "time_order" in kwargs:
            if not "time_order" in levelset_domain_local or levelset_domain_local["time_order"] == -1:
                levelset_domain_local["time_order"] = kwargs["time_order"]
            del kwargs["time_order"]

        # print("SymbolicBFI-Wrapper: SymbolicCutBFI called")
        return SymbolicCutBFI(levelset_domain=levelset_domain_local,*args, **kwargs)
    else:
        # print("SymbolicBFI-Wrapper: original SymbolicBFI called")
        if (levelset_domain == None):
            return ngsolve_SymbolicBFI(*args,**kwargs)
        else:
            return ngsolve_SymbolicBFI(levelset_domain,*args,**kwargs)

def SymbolicLFIWrapper(levelset_domain=None, *args, **kwargs):
    """
Wrapper around SymbolicLFI to allow for integrators on level set domains (see also
SymbolicCutLFI). The dictionary contains the level set function (CoefficientFunciton or
GridFunction) and the domain-type (NEG/POS/IF). If the dictionary is not provided, the standard
SymbolicLFI function from NGSolve will be called.

Parameters

levelset_domain : dictionary
  entries:
  * "levelset": 
    singe level set : ngsolve.CoefficientFunction
      CoefficientFunction that describes the geometry. In the best case lset is a GridFunction of an
      FESpace with scalar continuous piecewise (multi-) linear basis functions.
    multiple level sets: tuple(ngsolve.GridFunction)
      Tuple of GridFunctions that describe the geometry.
  * "domain_type" :
    single level set: {NEG,POS,IF} (ENUM) 
      Integration on the domain where either:
      * the level set function is negative (NEG)
      * the level set function is positive (POS)
      * the level set function is zero     (IF )
    multiple level sets: {tuple({ENUM}), list(tuple(ENUM)), DomainTypeArray}
      Integration on the domains specified
  * "subdivlvl" : int
    On simplex meshes a subtriangulation is created on which the level set function lset is
    interpolated piecewise linearly. Based on this approximation, the integration rule is
    constructed. Note: this argument only works on simplices.
  * "order" : int
    (default: entry does not exist or value -1)
    overwrites "order"-arguments in the integration
  * "quad_dir_policy" : {FIRST, OPTIMAL, FALLBACK} (ENUM)
    Integration direction policy for iterated integrals approach
    * first direction is used unless not applicable (FIRST)
    * best direction (in terms of transformation constant) is used (OPTIMAL)
    * subdivision into simplices is always used (FALLBACK)

Other Parameters :

  form : ngsolve.CoefficientFunction
    form to integrate

  VOL_or_BND : {VOL,BND}
    integrator is defined in the volume or boundary

  element_boundary : boolean
    Integration of the boundary of an element
    (not active for level set domains)

  skeleton : boolean
    Integration over element-interface

  definedon : Region
    Domain description on where the integrator is defined

  definedonelements: BitArray
    BitArray that allows integration only on elements or facets (if skeleton=True) that are marked
    True.

  deformation : GridFunction
      Specify a specific mesh deformation for a linear form

  order : int
    Modifies the order of the quadrature rule used. This is overruled by "order"-entry of the 
    levelset_domain dictionary, if the dictionary entry exists.

  time_order : int
    order in time that is used in the space-time integration. time_order=-1 means that no space-time
    rule will be applied. This is only relevant for space-time discretizations. Note that
    time_order can only be active if the key "time_order" of the levelset_domain is not set (or -1)
"""
    if levelset_domain != None and type(levelset_domain)==dict:
        # shallow copy is sufficient to modify "order" and "time_order" locally
        levelset_domain_local = levelset_domain.copy()      
        if "order" in kwargs:
            if not "order" in levelset_domain_local:
                levelset_domain_local["order"] = kwargs["order"]
            del kwargs["order"]
        if "time_order" in kwargs:
            if not "time_order" in levelset_domain_local or levelset_domain_local["time_order"] == -1:
                levelset_domain_local["time_order"] = kwargs["time_order"]
            del kwargs["time_order"]
        
        return SymbolicCutLFI(levelset_domain=levelset_domain_local,*args, **kwargs)
    else:
        if (levelset_domain == None):
            return ngsolve_SymbolicLFI(*args,**kwargs)
        else:
            return ngsolve_SymbolicLFI(levelset_domain,*args,**kwargs)

def Integrate_X_special_args(levelset_domain={}, cf=None, mesh=None, VOL_or_BND=VOL, order=5, time_order=-1, region_wise=False, element_wise = False, heapsize=1000000, ip_container=None):
    """
Integrate_X_special_args should not be called directly.
See documentation of Integrate.
    """
    levelset_domain_local = levelset_domain.copy() 
    if not "order" in levelset_domain_local or levelset_domain_local["order"] == -1:
        levelset_domain_local["order"] = order
    if not "time_order" in levelset_domain_local or levelset_domain_local["time_order"] == -1:
        levelset_domain_local["time_order"] = time_order
    return IntegrateX(levelset_domain = levelset_domain_local,
                      mesh=mesh, cf=cf,
                      ip_container=ip_container,
                      element_wise=element_wise,
                      heapsize=heapsize
                      )


##### THIS IS ANOTHER WRAPPER (original IntegrateX-interface is pretty ugly...) TODO
def Integrate(levelset_domain=None, *args, **kwargs):
    """
Integrate-wrapper. If a dictionary 'levelset_domain' is provided integration will be done on the
level set part of the mesh. The dictionary contains the level set function (CoefficientFunciton or
GridFunction) and the domain-type (NEG/POS/IF). If the dictionary is not provided, the standard
Integrate function from NGSolve will be called.

Parameters

levelset_domain : dictionary
  entries:
  * "levelset": 
    singe level set : ngsolve.CoefficientFunction
      CoefficientFunction that describes the geometry. In the best case lset is a GridFunction of an
      FESpace with scalar continuous piecewise (multi-) linear basis functions.
    multiple level sets: tuple(ngsolve.GridFunction)
      Tuple of GridFunctions that describe the geometry.
  * "domain_type" :
    single level set: {NEG,POS,IF} (ENUM) 
      Integration on the domain where either:
      * the level set function is negative (NEG)
      * the level set function is positive (POS)
      * the level set function is zero     (IF )
    multiple level sets: {tuple({ENUM}), list(tuple(ENUM)), DomainTypeArray}
      Integration on the domains specified
  * "subdivlvl" : int
    On simplex meshes a subtriangulation is created on which the level set function lset is
    interpolated piecewise linearly. Based on this approximation, the integration rule is
    constructed. Note: this argument only works on simplices.
  * "order" : int
    (default: entry does not exist or value -1)
    overwrites "order"-arguments in the integration (affects only spatial integration)
  * "time_order" : int 
    defines integration order in time (for space-time integrals only)
  * "quad_dir_policy" : {FIRST, OPTIMAL, FALLBACK} (ENUM)
    Integration direction policy for iterated integrals approach
    * first direction is used unless not applicable (FIRST)
    * best direction (in terms of transformation constant) is used (OPTIMAL)
    * subdivision into simplices is always used (FALLBACK)

mesh :
  Mesh to integrate on (on some part)

cf : ngsolve.CoefficientFunction
  the integrand

order : int (default = 5)
  integration order. Can be overruled by "order"-entry of the levelset_domain dictionary.

time_order : int (default = -1)
  integration order in time (for space-time integration), default: -1 (no space-time integrals)

region_wise : bool
  (only active for non-levelset version)

element_wise : bool
  integration result is return per element

ip_container : list (or None)
  a list to store integration points (for debugging or visualization purposes only!)

heapsize : int
  heapsize for local computations.
    """
    if levelset_domain != None and type(levelset_domain)==dict:
        # print("Integrate-Wrapper: IntegrateX called")
        return Integrate_X_special_args(levelset_domain, *args, **kwargs)
    else:
        # print("Integrate-Wrapper: original Integrate called")
        if (levelset_domain == None):
            return ngsolve_Integrate(*args,**kwargs)
        else:
            newargs = [levelset_domain]
            for q in args:
                newargs.append(q)
            return ngsolve_Integrate(*newargs,**kwargs)

def IndicatorCF(mesh, ba, facets = False):
    """
Returns a CoefficientFunction that evaluates a BitArray. On elements/facets with an index i where
the BitArray evaluates to true the CoefficientFunction will evaluate as 1, otherwise as 0. Similar
functionality (only on elements) can be obtained with BitArrayCF.
    """
    if facets:
        ret = GridFunction(FESpace("facet",mesh,order=0))
        for i in range(len(ba)):
            if ba[i]:
                ret.vec[i] = 1.0
            else:
                ret.vec[i] = 0.0
        return ret
    else:
        return BitArrayCF(BitArray(ba))

def CutRatioGF(cutinfo):
    """
Ratio between negative and full part of an element. Vector taken from CutInfo and put into a
piecewise constant GridFunction.
    """
    ret = GridFunction(L2(cutinfo.Mesh(),order=0))
    ret.vec.data = cutinfo.GetCutRatios(VOL)
    return ret

def kappa(mesh,lset_approx, subdivlvl=0):
    """
Tuple of ratios between negative/positive and full
part of an element (deprecated).
    """
    print("kappa-function is deprecated - use CutRatioGF instead")
    kappa1 = GridFunction(L2(mesh,order=0))
    lset_neg = { "levelset" : lset_approx, "domain_type" : NEG, "subdivlvl" : subdivlvl}
    kappa_f = LinearForm(kappa1.space)
    kappa_f += SymbolicLFI(levelset_domain = lset_neg, form = kappa1.space.TestFunction() )
    kappa_f.Assemble();
    kappa1.space.SolveM(CoefficientFunction(1.0),kappa_f.vec)
    kappa1.vec.data = kappa_f.vec
    kappa2 = 1.0 - kappa1
    return (kappa1,kappa2)

def IsCut(mesh,lset_approx, subdivlvl=0):
    """
GridFunction that is 1 on cut elements, 0 otherwise (deprecated). Use CutInfo-functionality (perhaps
combined with BitArrayCF).
    """
    print("IsCut-function is deprecated - use CutInfo-functionality instead")
    def cut(kappa):
        if (kappa > 1e-16 and kappa < 1.0-1e-16):
            return 1.0
        else:
            return 0.0
    kappa1, dummycf = kappa(mesh,lset_approx,subdivlvl)
    for i in range(len(kappa1.vec)):
        kappa1.vec[i] = cut(kappa1.vec[i])
    # from numpy import vectorize
    # cut = vectorize(cut)
    # kappa1.vec.FV().NumPy()[:] = cut(kappa1.vec.FV().NumPy())
    return kappa1

all_domain_types = [ DOMAIN_TYPE.NEG,
                     DOMAIN_TYPE.POS,
                     DOMAIN_TYPE.IF ]            

all_combined_domain_types = [ COMBINED_DOMAIN_TYPE.NO,
                              COMBINED_DOMAIN_TYPE.CDOM_NEG,
                              COMBINED_DOMAIN_TYPE.CDOM_POS,
                              COMBINED_DOMAIN_TYPE.UNCUT,
                              COMBINED_DOMAIN_TYPE.CDOM_IF,
                              COMBINED_DOMAIN_TYPE.HASNEG,
                              COMBINED_DOMAIN_TYPE.HASPOS,
                              COMBINED_DOMAIN_TYPE.ANY ]            

def SpaceTimeWeakSet(gfu_e, cf, space_fes):
    """
Ondocumented feature
    """
    gfu_e_repl = GridFunction(space_fes)
    gfu_e_repl.Set( cf )
    gfu_e.vec[:].data = gfu_e_repl.vec

ngsolveSet = GridFunction.Set
def SpaceTimeSet(self, cf, *args, **kwargs):
    """
Overrides the NGSolve version of Set in case of a space-time FESpace.
In this case the usual Set() is used on each nodal dof in time.
    """
    if (isinstance(self.space,CSpaceTimeFESpace)):
      cf = CoefficientFunction(cf)
      gfs = GridFunction(self.space.spaceFES)
      ndof_node = len(gfs.vec)
      j = 0
      for i,ti in enumerate(self.space.TimeFE_nodes()):
        if self.space.IsTimeNodeActive(i):
          ngsolveSet(gfs,fix_tref(cf,ti), *args, **kwargs)
          self.vec[j*ndof_node : (j+1)*ndof_node].data = gfs.vec[:]
          j += 1
    else:
      ngsolveSet(self,cf, *args, **kwargs)

def fix_tref(obj,time,*args,**kwargs):
    """
Takes a (possibly space-time) CoefficientFunction and fixes the temporal
variable to `time` and return this as a new CoefficientFunction.
Note that all operations are done on the unit interval it is the 
reference time that is fixed. 
    """

    if not isinstance(time, Parameter):
      if isinstance(obj,GridFunction) or isinstance(obj,ProxyFunction):
        if time == 0:
          return obj.Operator("fix_tref_bottom")
        elif time == 1: 
          return obj.Operator("fix_tref_top")
        elif isinstance(obj,GridFunction):
          return fix_tref_gf(obj,time,*args,**kwargs)
      elif isinstance(obj,ProxyFunction):
        return fix_tref_proxy(obj,time,*args,**kwargs)

    if isinstance(obj,CoefficientFunction):
      return fix_tref_coef(obj,time,*args,**kwargs)
    else:
      raise Exception("obj is not a CoefficientFunction")

def fix_t_coef(obj,time,*args,**kwargs):
  print("WARNING: fix_t_coef is deprecated. Use \"fix_tref_coef\" instead. \n         Note that operators act w.r.t. the reference time intervals.")
  return fix_tref_coef(obj,time,*args,**kwargs)
def fix_t_gf(obj,time,*args,**kwargs):
  print("WARNING: fix_t_gf is deprecated. Use \"fix_tref_gf\" instead. \n         Note that operators act w.r.t. the reference time intervals.")
  return fix_tref_gf(obj,time,*args,**kwargs)
def fix_t_proxy(obj,time,*args,**kwargs):
  print("WARNING: fix_t_proxy is deprecated. Use \"fix_tref_proxy\" instead. \n         Note that operators act w.r.t. the reference time intervals.")
  return fix_tref_proxy(obj,time,*args,**kwargs)


def fix_t(obj,time,*args,**kwargs):
  """
  Deprecated: use "fix_tref" instead
  """
  print("WARNING: fix_t is deprecated. Use \"fix_tref\" instead. \n         Note that operators act w.r.t. the reference time intervals.")
  return fix_tref(obj,time,*args,**kwargs)

import ngsolve
from ngsolve.internal import *


class DummyScene:
  def __init__(self):
    pass
  def Redraw(self,blocking=False):
    ngsolve.Redraw(blocking=blocking)
dummy_scene = DummyScene()


def DrawDiscontinuous_std(StdDraw,levelset, fneg, fpos, *args, **kwargs):
    def StdDrawWithDummyScene(cf,*args,**kwargs):
        StdDraw(cf,*args,**kwargs)
        return dummy_scene

    if "deformation" in kwargs and StdDraw.__module__ == "ngsolve.solve":
        args2 = list(args[:])
        args2[1] = "deformation_"+args[1]
        StdDraw(kwargs["deformation"],*tuple(args2),**kwargs)
        visoptions.deformation=1
    if not "sd" in kwargs:
        kwargs["sd"] = 5
        
    return StdDrawWithDummyScene(IfPos(levelset,fpos,fneg),*args,**kwargs)
    
def DrawDiscontinuous_webgui(WebGuiDraw,levelset, fneg, fpos, *args, **kwargs):
    fneg = CoefficientFunction(fneg)
    fpos = CoefficientFunction(fpos)
    if fneg.dim > 1 or fpos.dim > 1:
        print("webgui discontinuous vis only for scalar functions a.t.m., switching to IfPos variant")
    else:
        return WebGuiDraw(CoefficientFunction((levelset,fpos,fneg,0)),eval_function="value.x>0.0?value.y:value.z",*args,**kwargs)
    return DrawDiscontinuous_std(WebGuiDraw,levelset, fneg, fpos, *args, **kwargs)

from functools import partial
def MakeDiscontinuousDraw(Draw):
    """
Generates a Draw-like visualization function. If Draw is from the webgui, a special evaluator is used to draw a pixel-sharp discontinuity otherwise an IfPos-CoefficientFunction is used.     
    """
    if (Draw.__module__ == "ngsolve.webgui"):
        ret = partial(DrawDiscontinuous_webgui,Draw)
    else:
        ret = partial(DrawDiscontinuous_std,Draw)
    ret.__doc__ ="""
    Visualization method for functions that are non-smooth across 
    level set interfaces. Effectively calls """+Draw.__module__+""".Draw with
    a few manipulations. 

    Parameters
    ----------
    levelset : CoefficientFunction
        (scalar) CoefficientFunction that describes the (implicit) geometry 
    fneg : CoefficientFunction
        CoefficientFunction that is evaluated where the level set function is negative
    fpos : CoefficientFunction
        CoefficientFunction that is evaluated where the level set function is positive
        
    deformation : deformation (optional)
        (vectorial) CoefficientFunction that describes the deformation of the background mesh
    *remainder* : *
        all remainder arguments are passed to """ +Draw.__module__ +".Draw"
    return ret

class NoDeformation:
    """
Dummy deformation class. Does nothing to the mesh. Has two dummy members:
  * lset_p1 : ngsolve.GridFunction
    The piecewise linear level set function 
  * deform : ngsolve.GridFunction
    A zero GridFunction (for compatibility with netgen Draw(... deformation=)) 
    """

    def __init__(self, mesh=None, levelset=None):
        self.deform = GridFunction(H1(mesh, order=1, dim=mesh.dim), "dummy_deform")
        self.deform.vec.data[:] = 0.0
        if levelset != None:
            if mesh == None:
                raise Exception("need mesh")
            self.lset_p1 = GridFunction(H1(mesh, order=1))
            InterpolateToP1(levelset, self.lset_p1)

    def __enter__(self):
        return self.lset_p1

    def __exit__(self, type, value, tb):
        pass


try:
    __IPYTHON__
    from ipywidgets import interact, FloatSlider
    from ngsolve.webgui import Draw
    def TimeSlider_Draw(cf,mesh,*args,**kwargs):
        ts = Parameter(0)
        if not isinstance(cf,CoefficientFunction):
            cf = CoefficientFunction(cf)
        scene = Draw(fix_tref(cf,ts),mesh,*args,**kwargs); 
        def UpdateTime(time): 
            ts.Set(time); scene.Redraw()
        return interact(UpdateTime,time=FloatSlider(description="tref:", 
                                                    continuous_update=False,
                                                    min=0,max=1,step=.025))
    def TimeSlider_DrawDC(cf1,cf2,cf3,mesh,*args,**kwargs):
        """
Draw a (reference) time-dependent function that is discontinuous across an 
interface described by a level set function. Change reference time through
widget slider.

        Args:
            cf1 (CoefficientFunction): level set function
            cf2 (CoefficientFunction): function to draw where lset is negative
            cf3 (CoefficientFunction): function to draw where lset is positive
            mesh (Mesh): Mesh

        Returns:
            widget element that allows to vary the reference time. 
        """
        DrawDC = MakeDiscontinuousDraw(Draw)
        if not isinstance(cf1,CoefficientFunction):
            cf1=CoefficientFunction(cf1)
        if not isinstance(cf2,CoefficientFunction):
            cf2=CoefficientFunction(cf2)
        if not isinstance(cf3,CoefficientFunction):
            cf3=CoefficientFunction(cf3)
        ts = Parameter(0)
        scene = DrawDC(fix_tref(cf1,ts),fix_tref(cf2,ts),fix_tref(cf3,ts),mesh,*args,**kwargs); 
        def UpdateTime(time): 
            ts.Set(time); scene.Redraw()
        return interact(UpdateTime,time=FloatSlider(description="tref:", 
                                                    continuous_update=False,
                                                    min=0,max=1,step=.025))
    import ngsolve.webgui
    DrawDC = MakeDiscontinuousDraw(ngsolve.webgui.Draw)
        
except:
    def TimeSlider_Draw(cf,mesh,*args,**kwargs):
      print("TimeSlider_Draw only available in ipython mode")
    def TimeSlider_DrawDC(cf1,cf2,cf3,mesh,*args,**kwargs):
      print("TimeSlider_DrawDC only available in ipython mode")
    import ngsolve
    DrawDC = MakeDiscontinuousDraw(ngsolve.Draw)


_dCut_raw = CutDifferentialSymbol(VOL)
_dFacetPatch_raw = FacetPatchDifferentialSymbol(VOL)

def dFacetPatch(**kwargs):
    """
    Differential symbol for facet patch integrators.

    Parameters
    ----------
    definedon : Region
        Domain description on where the integrator is defined.
    deformation : ngsolve.GridFunction
        Mesh deformation that is applied during integration. Default: None.
    definedonelements : ngsolve.BitArray
        Allows integration only on a set of facets
        that are marked True. Default: None.
    time_order : int
        Order in time that is used in the space-time integration.
        Default: time_order=-1 means that no space-time rule will be
        applied. This is only relevant for space-time discretizations.
    tref : double
        Turn spatial integration into space-time integration with 
        fixed time tref.

    Returns
    -------
      FacetPatchDifferentialSymbol(VOL)
    """
    if "element_vb" in kwargs or "element_boundary" in kwargs \
       or "skeleton" in kwargs:
        raise Exception("facet patch integrators are fixed to facet patches")
    return _dFacetPatch_raw(**kwargs)


def dCut(levelset, domain_type, order=None, subdivlvl=None, time_order=-1,
         levelset_domain=None, **kwargs):
    """
    Differential symbol for cut integration.

    Parameters
    ----------
    levelset : ngsolve.GridFunction
        The level set fct. describing the geometry 
        (desirable: P1 approximation).
    domain_type : {POS, IF, NEG, mlset.DomainTypeArray}
        The domain type of interest.
    order : int
        Modify the order of the integration rule used.
    subdivlvl : int
        Number of additional subdivision used on cut elements to
        generate the cut quadrature rule. Note: subdivlvl >0 only
        makes sense if you don't provide a P1 level set function
        and no isoparametric mapping is used.
    definedon : Region
        Domain description on where the integrator is defined.
    element_boundary : bool
        Integration on each element boundary. Default: False
    element_vb : {VOL, BND, BBND}
        Integration on each element or its (B)boundary. Default: VOL
        (is overwritten by element_boundary if element_boundary 
        is True)
    skeleton : bool
        Integration over element-interface. Default: False.
    deformation : ngsolve.GridFunction
        Mesh deformation that is applied. Default: None.
    definedonelements : ngsolve.BitArray
        Allows integration only on elements or facets (if skeleton=True)
        that are marked True. Default: None.
    time_order : int
        Order in time that is used in the space-time integration.
        Default: time_order=-1 means that no space-time rule will be
        applied. This is only relevant for space-time discretizations.
    tref : float
        turns a spatial integral resulting in spatial integration rules
        into a space-time quadrature rule with fixed reference time tref
    levelset_domain : dict
        description of integration domain through a dictionary 
        (deprecated).

    Returns
    -------
        CutDifferentialSymbol(VOL)
    """
    if levelset_domain is not None and type(levelset_domain) == dict:
        lsetdom = levelset_domain
    else:
        lsetdom = {"levelset": levelset, "domain_type": domain_type}
    if order is not None and "order" not in lsetdom.keys():
        lsetdom["order"] = order
    if subdivlvl is not None and "subdivlvl" not in lsetdom.keys():
        lsetdom["subdivlvl"] = subdivlvl
    if time_order > -1 and "time_order" not in lsetdom.keys():
        lsetdom["time_order"] = time_order
    if "tref" in kwargs:
        lsetdom["tref"] = kwargs["tref"]
        del kwargs["tref"]

    return _dCut_raw(lsetdom, **kwargs)


def dxtref(mesh, order=None, time_order=-1, **kwargs):
    """
    Differential symbol for the integration over all elements extruded by
    the reference interval [0,1] to space-time prisms.

    Parameters
    ----------
    mesh : ngsolve.Mesh
        The spatial mesh.
        The domain type of interest.
    order : int
        Modify the order of the integration rule used.
    definedon : Region
        Domain description on where the integrator is defined.
    element_boundary : bool
        Integration on each element boundary. Default: False
    element_vb : {VOL, BND, BBND}
        Integration on each element or its (B)boundary. Default: VOL
        (is overwritten by element_boundary if element_boundary 
        is True)
    skeleton : bool
        Integration over element-interface. Default: False.
    deformation : ngsolve.GridFunction
        Mesh deformation. Default: None.
    definedonelements : ngsolve.BitArray
        Allows integration only on elements or facets (if skeleton=True)
        that are marked True. Default: None.
    time_order : int
        Order in time that is used in the space-time integration.
        Default: time_order=-1 means that no space-time rule will be
        applied. This is only relevant for space-time discretizations.

    Return
    ------
        CutDifferentialSymbol(VOL)
    """
    gflset = GridFunction(H1(mesh))
    gflset.vec[:] = 1

    lsetdom = {"levelset": gflset, "domain_type": POS}
    if order is not None:
        lsetdom["order"] = order
    if time_order > -1:
        lsetdom["time_order"] = time_order

    return _dCut_raw(lsetdom, **kwargs)

def dmesh(mesh=None,*args,**kwargs):
    """
    Differential symbol for the integration over all elements in the mesh.

    Parameters
    ----------
    mesh : ngsolve.Mesh
        The spatial mesh.
        The domain type of interest.
    definedon : Region
        Domain description on where the integrator is defined.
    element_boundary : bool
        Integration on each element boundary. Default: False
    element_vb : {VOL, BND, BBND}
        Integration on each element or its (B)boundary. Default: VOL
        (is overwritten by element_boundary if element_boundary 
        is True)
    skeleton : bool
        Integration over element-interface. Default: False.
    deformation : ngsolve.GridFunction
        Mesh deformation. Default: None.
    definedonelements : ngsolve.BitArray
        Allows integration only on elements or facets (if skeleton=True)
        that are marked True. Default: None.
    tref : float
        turns a spatial integral resulting in spatial integration rules
        into a space-time quadrature rule with fixed reference time tref

    Return
    ------
        CutDifferentialSymbol(VOL)
    """
    if "tref" in kwargs:
        if mesh == None:
            raise Exception("dx(..,tref..) needs mesh")
        gflset = GridFunction(H1(mesh))
        gflset.vec[:] = 1
        lsetdom = {"levelset": gflset, "domain_type": POS, "tref" : kwargs["tref"]}
        del kwargs["tref"]
        return _dCut_raw(lsetdom, **kwargs)
    else:
        return dx(*args,**kwargs)

def RestrictedBilinearForm(space=None,name="blf",element_restriction=None,facet_restriction=None,trialspace=None,testspace=None,**kwargs):
    """
    Creates a restricted bilinear form, which is bilinear form with a reduced MatrixGraph
    compared to the usual BilinearForm. BitArray(s) define on which elements/facets entries will be
    created.

    Use cases:

      * ghost penalty type stabilization:
        Facet-stabilization that are introduced only act on a few facets in the mesh. By providing the
        information on the corresponding facets, these additional couplings will only be introduced
        where necessary.

      * fictitious domain methods (domain decomposition methods):
        When PDE problems are only solved on a part of a domain while a finite element space is used
        that is still defined on the whole domain, a BitArray can be used to mark the 'active' part of
        the mesh. 

    Parameters

    space: ngsolve.FESpace
      finite element space on which the bilinear form is defined. If trial space and test space are different 
      they can be specified using the trialspace and testspace arguments.

    name : string
      name of the bilinear form

    element_restriction : ngsolve.BitArray
      BitArray defining the 'active mesh' element-wise

    facet_restriction : ngsolve.BitArray
      BitArray defining the 'active facets'. This is only relevant if FESpace has DG-terms (dgjumps=True)
    
    trialspace : ngsolve.FESpace
      finite element space on which the bilinear form is defined
      (trial space).
    
    testspace : ngsolve.FESpace
      finite element space on which the bilinear form is defined
      (test space).

    kwargs : keyword arguments
      kwargs are pasre to flags and passed to bilinearform 
    """

    argument_list = [name]
    if element_restriction != None:
        argument_list.append(element_restriction)
    if facet_restriction != None:
        argument_list.append(facet_restriction)

    if trialspace != None and testspace != None:
        if trialspace.is_complex and testspace.is_complex:
            return RestrictedBilinearFormComplex(trialspace,testspace,*argument_list,**kwargs)
        elif (not trialspace.is_complex) and (not testspace.is_complex):
            return RestrictedBilinearFormDouble(trialspace,testspace,*argument_list,**kwargs)
        else:
            raise Exception("Trialspace is real and testspace is complex or vice versa. RestrictedBilinearForm not implemented for this case!")
    else: 
        if space == None:
            raise Exception("No space given for RestrictedBilinearForm!")
        if space.is_complex:
            return RestrictedBilinearFormComplex(space,*argument_list,**kwargs)
        else:
            return RestrictedBilinearFormDouble(space,*argument_list,**kwargs)


# some global scope manipulations (monkey patches etc..):

# monkey patches
# print("|---------------------------------------------|")
# print("| ngsxfem applied monkey patches for          |")
# print("| GridFunction.Set, SymbolicLFI, SymbolicBFI. |")
# print("|---------------------------------------------|")
GridFunction.Set = SpaceTimeSet
SymbolicLFI = SymbolicLFIWrapper
SymbolicBFI = SymbolicBFIWrapper

# global scope definitions:
tref = ReferenceTimeVariable()


from ngsolve.fem import *
from ngsolve.comp import *
from ngsolve.solve import *
from ngsolve.la import *
from ngsolve.utils import *

from xfem import *

class LevelSetMeshAdaptation:
    """
Class to compute a proper mesh deformation to improve a piecewise (multi-) linear level set
approximation to obtain a higher order accurate approximation.

Computes a continuous function that describes a shift between points that are on the (P1 )
approximated level set function and its higher order accurate approximation. The transformation is
only applied on elements where a level value inside a certain interval (lower,upper) exists.

The result is a deform function (D) which is computed pointwise as

1)phi_lin( Psi(x) ) = phi_h(x)

  with Psi(x) = x + d(x) qn(x) =: D(x)

for all x on 'cut' elements

with

  phi_h : self.lset_ho
    the higher order level set function

  phi_lin : self.lset_p1
    the P1 level set function

  Psi : Id + self.deform
    the resulting deformation

  qn : self.qn
    normal direction field

This class holds its own members for the higher order an lower order (P1) approximation of the level
set function and only depends on the input of a mesh and a CoefficientFunction (the levelset
function) (and options).

A LevelSetMeshAdaptation can also be used as a context  manager. In this case, 
the mesh deformation is applied on the mesh inside the context.
    """

    order_deform = 2
    order_qn = 2
    order_lset = 2

    @TimeFunction
    def __init__(self, mesh, order = 2, lset_lower_bound = 0, lset_upper_bound = 0, 
                 threshold = -1, discontinuous_qn = False, 
                 eps_perturbation = 1e-14, heapsize=1000000,
                 levelset = None):
        """
The computed deformation depends on different options:

  order : int
    order of deformation GridFunction (ideally) order + 1 is the accuracy of the geometry
    approximation after applying the deformation on the mesh

  lset_lower_bound: float
    smallest relevant level set value to define the 'cut' elements where the mapping should be
    applied
    
  lset_upper_bound: float
    largest relevant level set value to define the 'cut' elements where the mapping should be
    applied
    
  threshold: float
    maximum (pointwise) value for d(x)/h in the mapping
      Psi(x) = x + d(x) qn(x)
    of the mesh transformation. A small value might be necessary if the geometry is only coarsely
    approximated to avoid irregular meshes after a corresponding mesh deformation.

  discontinuous_qn: boolean
    As an approximation for the normal direction we use n_h = nabla phi_h (gradient of higher order
    level set approximation) depending on the discontinuous_qn flag this normal field will be
    projected onto a continuous finite element space or not.

  eps_perturbation: float
    epsilon perturbation that is used to interpolate to P1

  heapsize : int
    heapsize for local computations.

  levelset : CoefficientFunction or None(default)
    If a level set function is prescribed the deformation is computed right away.
        """
        self.order_deform = order
        self.order_qn = order
        self.order_lset = order

        self.lset_lower_bound = lset_lower_bound
        self.lset_upper_bound = lset_upper_bound
        self.threshold = threshold

        self.eps_perturbation = eps_perturbation
        
        self.mesh = mesh
        self.v_ho = H1(mesh, order=self.order_lset)
        self.lset_ho = GridFunction (self.v_ho, "lset_ho")

        if (discontinuous_qn):
            self.v_qn = L2(mesh, order=self.order_qn, dim=mesh.dim)
        else:
            self.v_qn = H1(mesh, order=self.order_qn, dim=mesh.dim)
        self.qn = GridFunction(self.v_qn, "qn")
    
        self.v_p1 = H1(mesh, order=1)
        self.lset_p1 = GridFunction (self.v_p1, "lset_p1")

        self.v_def = H1(mesh, order=self.order_deform, dim=mesh.dim)
        self.deform = GridFunction(self.v_def, "deform")
        self.deform_last = GridFunction(self.v_def, "deform_last")
        self.heapsize = heapsize
        self.gf_to_project = []
        if levelset != None:
          self.CalcDeformation(levelset)

        
    def ProjectOnUpdate(self,gf,update_domain=None):
        """
When the LevelsetMeshAdaptation class generates a new deformation (due to 
a new level set function) all GridFunction that have been stored through
`ProjectOnUpdate` will be projected from the previous to the new mesh
by an essentially local projection (Oswald projection of the shifted 
evaluation)

  gfu : GridFunction (or list of GridFunctions)
    GridFunction(s) to store for later deformation updates.
        """      
        if isinstance(gf,list):
            self.gf_to_project.extend((gf, update_domain))
        else:
            self.gf_to_project.append((gf, update_domain))

    @TimeFunction
    def ProjectGFs(self):
        """
ProjectGFs projects all stored GridFunctions to the currect deformation.
This function is typically only called in the `CalcDeformation` unless
`CalcDeformation` is called with the argument `dont_project_gfs = True`.
        """      

        timer = Timer("ProjectGFs")
        timer.Start()
        for (gf, update_domain) in self.gf_to_project:
            # make tmp copy 
            gfcopy = GridFunction(gf.space)
            gfcopy.vec.data = gf.vec
            if update_domain is None:
                gf.Set(shifted_eval(gfcopy, back = self.deform_last, forth = self.deform))
            else:
                gf.Set(shifted_eval(gfcopy, back = self.deform_last, forth = self.deform), definedonelements=update_domain)
            #print("updated ", gf.name)
        timer.Stop()

    @TimeFunction
    def CalcDeformation(self, levelset, ba =None, blending=None, dont_project_gfs=False):
        """
Compute the mesh deformation, s.t. isolines on cut elements of lset_p1 (the piecewise linear
approximation) are mapped towards the corresponding isolines of a given function

Parameters:

levelset : CoefficientFunction
  The coefficient function that prescribes where the piecewise linear iso lines should be mapped to.

blending : None/string/CoefficientFunction
  Option to apply the mesh deformation more localized on cut elements. Setting blending function to
  0 (CoefficientFunction(0.0) or None or "none") corresponds to applying the mapping on all points
  on cut elements completely. Using a blending function as a CoefficientFunction allows for a
  transition between the full application of the mapping (value 0) and no application of the mapping
  (value 1). There are predefined blending functions that are selectable via a string:
   * "quadratic":
     blending function that is 0 at the zero level set (of lset_p1) and increases quadratically with
     lset_p1. It is scaled with h, so that value 1 is not reached within cut elements.
   * "quartic":
     blending function that is 0 at the zero level set (of lset_p1) and increases like a fourth
     order polynomial with lset_p1. It is scaled with h, so that value 1 is not reached within cut
     elements.
dont_project_gfs : bool
  Usually, all stored GridFunctions are projected on the new deformed mesh. This flag
  can be used to avoid the call of the projection step (`ProjectGFs`).
        """
        self.v_ho.Update()
        self.lset_ho.Update()
        self.v_p1.Update()
        self.lset_p1.Update()
        self.v_qn.Update()
        self.qn.Update()
        self.v_def.Update()
        self.deform.Update()
        self.deform_last.Update()
        
        self.lset_ho.Set(levelset)
        self.qn.Set(self.lset_ho.Deriv())
        InterpolateToP1(self.lset_ho,self.lset_p1,eps_perturbation=self.eps_perturbation)
        if blending == None or blending == "none":
            blending = CoefficientFunction(0.0)
        elif blending == "quadratic":
            scale=sqrt(self.lset_p1.space.mesh.dim) * specialcf.mesh_size
            blending = self.lset_p1*self.lset_p1/( scale * scale)
        elif blending == "quartic":
            scale=sqrt(self.lset_p1.space.mesh.dim) * specialcf.mesh_size
            blending = self.lset_p1*self.lset_p1*self.lset_p1*self.lset_p1/(scale*scale*scale*scale)
            
        self.deform_last.vec.data = self.deform.vec
        ProjectShift(self.lset_ho,
                     self.lset_p1,
                     self.deform,
                     self.qn,
                     ba,
                     blending,
                     lower=self.lset_lower_bound,
                     upper=self.lset_upper_bound,
                     threshold=self.threshold,
                     heapsize=self.heapsize)

        if not dont_project_gfs:
            self.ProjectGFs()

        return self.deform

    def __enter__(self):
      self.mesh.SetDeformation(self.deform)
      return self.lset_p1

    def __exit__(self, type, value, tb):
      self.mesh.UnsetDeformation()

    @TimeFunction
    def CalcMaxDistance(self, levelset, order=None, heapsize=None, deform=False):
        """
Compute approximated distance between of the isoparametrically obtained geometry
(should be called in deformed state)
        """
        if order == None:
          order = 2 * self.order_qn
        if heapsize == None:
          heapsize = self.heapsize
        lset_dom = {"levelset": self.lset_p1, "domain_type" : IF, "order": order}
        if self.mesh.deformation or not deform:
          minv, maxv = IntegrationPointExtrema(lset_dom, self.mesh, levelset, heapsize=heapsize)
        else:
          self.mesh.deformation = self.deform
          minv, maxv = IntegrationPointExtrema(lset_dom, self.mesh, levelset, heapsize=heapsize)
          self.mesh.deformation = None
        return max(abs(minv),abs(maxv))

    def levelset_domain(self, domain_type = IF):
        """
  Return a levelset_domain dictionary.
        Args:
            domain_type (DOMAIN_TYPE, optional): domain type for level set doamin. Defaults to IF.

        Returns:
            dict: levelset domain
        """
        return { "levelset" : self.lset_p1, "domain_type" : domain_type}

    def Integrator(self, SymbolicFI, domain_type, form, definedonelements = None):
        """
  Convenience function to construct Symbolic(Cut)LFI/Symbolic(Cut)BFI
        """    
        fi = SymbolicFI(levelset_domain = self.levelset_domain(domain_type), 
                         form = form, 
                         deformation=self.deform)
        if definedonelements != None:
            fi.SetDefinedOnElements(definedonelements)       
        return fi

    def LFI(self, domain_type, form, definedonelements = None):
        """
  Convenience function to construct Symbolic(Cut)LFI based on the domain_type 
  and the known level set function.
        """    
        return self.Integrator(SymbolicLFI, domain_type, form, definedonelements)

    def BFI(self, domain_type, form, definedonelements = None):
        """
  Convenience function to construct Symbolic(Cut)BFI based on the domain_type
  and the known level set function.
        """    
        return self.Integrator(SymbolicBFI, domain_type, form, definedonelements)

    @TimeFunction
    def Integrate(self, domain_type, cf, order = 5):
        """
  Convenience function to Integrate on cut domain.
        """    
        self.mesh.SetDeformation(self.deform)
        fi = Integrate(levelset_domain = self.levelset_domain(domain_type), 
                       mesh = self.mesh, cf = cf, order = order)
        self.mesh.UnsetDeformation()
        return fi


    @TimeFunction
    def MarkForRefinement(self, levelset = None, refine_threshold = 0.1, absolute = False):
        """
Marks elements for refinement where the geometry approximation is larger than a prescrbed relative
(or absolute) (to the mesh size) error. This will lead to a refinement in zones with high curvature.

Parameters:

  levelset : ngsolve.CoefficientFunction/None
    accurate level set function that is used to compute the error. If None self.lset_ho will be
    used.

  refine_threshold : float
    value that decides if a geometry error is large enough to mark for refinement or not

  absolute : boolean
    decides if the refine_threshold is an absolute value or if it is weighted with the mesh size
        """
        lset_stats = StatisticContainer()
        if levelset==None:
            levelset = self.lset_ho
        CalcDistances(lset_ho=levelset,lset_p1=self.lset_p1,deform=self.deform,stats=lset_stats,refine_threshold=refine_threshold,absolute=absolute)
        
#     def CalcDeformationError(self, deform_stats):
#         """
# Compute error between ideal and realized deformation
#         """
#         CalcDeformationError(self.lset_ho,self.lset_p1,self.deform,deform_stats,self.lset_lower_bound, self.lset_upper_bound)
        



__ngsolve_required__ = "6.2.2105-0"

def check_if_ngsolve_newer_than(ngsolve_version_required):
    """
    Check for compatibility of ngsolve version
    """
    import ngsolve
    import re
    ngsver = [0 for i in range(4)]
    res = re.split(r'-', ngsolve.__version__)
    if len(res) > 1:
        mmp, ngsver[3], h = res
    else:
        mmp = res[0]
        ngsver[3] = "0"
    ngsver[0], ngsver[1], ngsver[2] = re.split(re.compile('[.bv]'), mmp)

    ngsver_required = [0 for i in range(4)]
    mmp, ngsver_required[3] =  re.split(re.compile('[-bv]'), ngsolve_version_required)
    ngsver_required[0:3] = re.split(re.compile('[.bv]'), mmp)

    ret = True
    for i in range(4):
        if int(ngsver[i]) < int(ngsver_required[i]):
            ret = False
            print("""
####################################################
                     WARNING!""")
            print("Your NGSolve-Version (" + ngsolve.__version__+")")
            print("is older than required (>=" + ngsolve_version_required +").",end="")
            print("""
####################################################
""")
            return False
        elif int(ngsver[i]) > int(ngsver_required[i]):
            break
    return ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
import sys
import platform
import subprocess
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion

class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)
class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " +
                               ", ".join(e.name for e in self.extensions))
        if platform.system() == "Windows":
            cmake_version = LooseVersion(re.search(r'version\s*([\d.]+)', out.decode()).group(1))
            if cmake_version < '3.1.0':
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")
        for ext in self.extensions:
            self.build_extension(ext)
    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        # required for auto-detection of auxiliary "native" libs
        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable,
                      '-DCMAKE_CXX_COMPILER=ngscxx',
                      '-DCMAKE_LINKER=ngsld',
                      '-DBUILD_STUB_FILES=ON',
                      '-DBUILD_NGSOLVE=OFF']
        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]
        if platform.system() == "Windows":
            #not expected to work... (but who knows..)
            cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)]
            if sys.maxsize > 2**32:
                cmake_args += ['-A', 'x64']
            build_args += ['--', '/m']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            build_args += ['--', '-j2']
        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=self.build_temp)
        
        subprocess.check_call(['mv', 'ngsxfem_py.so', 'xfem'], cwd=self.build_lib)

setup(
    name='xfem',
    version='2.0.2105',
    author='Christoph Lehrenfeld',
    author_email='lehrenfeld@math.uni-goettingen.de',
    description='(ngs)xfem is an Add-on library to Netgen/NGSolve for unfitted/cut FEM.',
    long_description='(ngs)xfem is an Add-on library to Netgen/NGSolve which enables the use of unfitted finite element technologies known as XFEM, CutFEM, TraceFEM, Finite Cell, ... . ngsxfem is an academic software. Its primary intention is to facilitate the development and validation of new numerical methods.',
    url="https://github.com/ngsxfem/ngsxfem",
    ext_modules=[CMakeExtension('ngsxfem_py')],
    cmdclass=dict(build_ext=CMakeBuild),
    packages=["xfem"],
    package_dir={"xfem": "python"},
    python_requires='>=3.5',
)
